@@ -28,6 +28,9 @@
 use std::mem;
 use std::ptr;
 
+/// Maximum value that can be encoded via varint.
+pub const MAX_VAR_INT: u64 = 4_611_686_018_427_387_903;
+
 /// A specialized [`Result`] type for [`OctetsMut`] operations.
 ///
 /// [`Result`]: https://doc.rust-lang.org/std/result/enum.Result.html
@@ -37,7 +40,7 @@ pub type Result<T> = std::result::Result<T, BufferTooShortError>;
 /// An error indicating that the provided [`OctetsMut`] is not big enough.
 ///
 /// [`OctetsMut`]: struct.OctetsMut.html
-#[derive(Clone, Copy, Debug, PartialEq)]
+#[derive(Clone, Copy, Debug, PartialEq, Eq)]
 pub struct BufferTooShortError;
 
 impl std::fmt::Display for BufferTooShortError {
@@ -52,6 +55,14 @@ impl std::error::Error for BufferTooShortError {
     }
 }
 
+/// Helper macro that asserts at compile time. It requires that
+/// `cond` is a const expression.
+macro_rules! static_assert {
+    ($cond:expr) => {{
+        const _: () = assert!($cond);
+    }};
+}
+
 macro_rules! peek_u {
     ($b:expr, $ty:ty, $len:expr) => {{
         let len = $len;
@@ -61,6 +72,7 @@ macro_rules! peek_u {
             return Err(BufferTooShortError);
         }
 
+        static_assert!($len <= mem::size_of::<$ty>());
         let mut out: $ty = 0;
         unsafe {
             let dst = &mut out as *mut $ty as *mut u8;
@@ -95,6 +107,7 @@ macro_rules! put_u {
 
         let dst = &mut $b.buf[$b.off..($b.off + len)];
 
+        static_assert!($len <= mem::size_of::<$ty>());
         unsafe {
             let src = &<$ty>::to_be(v) as *const $ty as *const u8;
             let off = (mem::size_of::<$ty>() - len) as isize;
@@ -118,7 +131,7 @@ macro_rules! put_u {
 /// Additionally, an offset (initially set to the start of the buffer) is
 /// incremented as bytes are read from / written to the buffer, to allow for
 /// sequential operations.
-#[derive(Debug, PartialEq)]
+#[derive(Debug, PartialEq, Eq)]
 pub struct Octets<'a> {
     buf: &'a [u8],
     off: usize,
@@ -127,8 +140,8 @@ pub struct Octets<'a> {
 impl<'a> Octets<'a> {
     /// Creates an `Octets` from the given slice, without copying.
     ///
-    /// Since there's no copy, the input slice needs to be mutable to allow
-    /// modifications.
+    /// Since the `Octets` is immutable, the input slice needs to be
+    /// immutable.
     pub fn with_slice(buf: &'a [u8]) -> Self {
         Octets { buf, off: 0 }
     }
@@ -197,7 +210,7 @@ impl<'a> Octets<'a> {
 
     /// Reads `len` bytes from the current offset without copying and advances
     /// the buffer.
-    pub fn get_bytes(&mut self, len: usize) -> Result<Octets> {
+    pub fn get_bytes(&mut self, len: usize) -> Result<Octets<'a>> {
         if self.cap() < len {
             return Err(BufferTooShortError);
         }
@@ -214,7 +227,7 @@ impl<'a> Octets<'a> {
 
     /// Reads `len` bytes from the current offset without copying and advances
     /// the buffer, where `len` is an unsigned 8-bit integer prefix.
-    pub fn get_bytes_with_u8_length(&mut self) -> Result<Octets> {
+    pub fn get_bytes_with_u8_length(&mut self) -> Result<Octets<'a>> {
         let len = self.get_u8()?;
         self.get_bytes(len as usize)
     }
@@ -222,7 +235,7 @@ impl<'a> Octets<'a> {
     /// Reads `len` bytes from the current offset without copying and advances
     /// the buffer, where `len` is an unsigned 16-bit integer prefix in network
     /// byte-order.
-    pub fn get_bytes_with_u16_length(&mut self) -> Result<Octets> {
+    pub fn get_bytes_with_u16_length(&mut self) -> Result<Octets<'a>> {
         let len = self.get_u16()?;
         self.get_bytes(len as usize)
     }
@@ -230,14 +243,57 @@ impl<'a> Octets<'a> {
     /// Reads `len` bytes from the current offset without copying and advances
     /// the buffer, where `len` is an unsigned variable-length integer prefix
     /// in network byte-order.
-    pub fn get_bytes_with_varint_length(&mut self) -> Result<Octets> {
+    pub fn get_bytes_with_varint_length(&mut self) -> Result<Octets<'a>> {
         let len = self.get_varint()?;
         self.get_bytes(len as usize)
     }
 
+    /// Decodes a Huffman-encoded value from the current offset.
+    ///
+    /// The Huffman code implemented is the one defined for HPACK (RFC7541).
+    #[cfg(feature = "huffman_hpack")]
+    pub fn get_huffman_decoded(&mut self) -> Result<Vec<u8>> {
+        use self::huffman_table::DECODE_TABLE;
+
+        const FLAG_END: u8 = 1;
+        const FLAG_SYM: u8 = 2;
+        const FLAG_ERR: u8 = 4;
+
+        // Max compression ratio is >= 0.5.
+        let mut out = Vec::with_capacity(self.cap() << 1);
+
+        let mut state = 0;
+        let mut eos = false;
+
+        while let Ok(byte) = self.get_u8() {
+            for data in [byte >> 4, byte & 0xf] {
+                let (next, sym, flags) = DECODE_TABLE[state][(data) as usize];
+
+                if flags & FLAG_ERR == FLAG_ERR {
+                    // Data followed the "end" marker.
+                    return Err(BufferTooShortError);
+                } else if flags & FLAG_SYM == FLAG_SYM {
+                    out.push(sym);
+                }
+
+                state = next;
+
+                // `eos` only correct when handling the byte & 0xf case; ignored
+                // and overwritten in the byte >> 4 case.
+                eos = flags & FLAG_END == FLAG_END;
+            }
+        }
+
+        if state != 0 && !eos {
+            return Err(BufferTooShortError);
+        }
+
+        Ok(out)
+    }
+
     /// Reads `len` bytes from the current offset without copying and without
     /// advancing the buffer.
-    pub fn peek_bytes(&self, len: usize) -> Result<Octets> {
+    pub fn peek_bytes(&self, len: usize) -> Result<Octets<'a>> {
         if self.cap() < len {
             return Err(BufferTooShortError);
         }
@@ -250,8 +306,19 @@ impl<'a> Octets<'a> {
         Ok(out)
     }
 
+    /// Rewinds the buffer offset by `len` elements.
+    pub fn rewind(&mut self, len: usize) -> Result<()> {
+        if self.off() < len {
+            return Err(BufferTooShortError);
+        }
+
+        self.off -= len;
+
+        Ok(())
+    }
+
     /// Returns a slice of `len` elements from the current offset.
-    pub fn slice(&'a self, len: usize) -> Result<&'a [u8]> {
+    pub fn slice(&self, len: usize) -> Result<&'a [u8]> {
         if len > self.cap() {
             return Err(BufferTooShortError);
         }
@@ -260,7 +327,7 @@ impl<'a> Octets<'a> {
     }
 
     /// Returns a slice of `len` elements from the end of the buffer.
-    pub fn slice_last(&'a self, len: usize) -> Result<&'a [u8]> {
+    pub fn slice_last(&self, len: usize) -> Result<&'a [u8]> {
         if len > self.cap() {
             return Err(BufferTooShortError);
         }
@@ -290,13 +357,18 @@ impl<'a> Octets<'a> {
         self.buf.len()
     }
 
+    /// Returns `true` if the buffer is empty.
+    pub fn is_empty(&self) -> bool {
+        self.buf.len() == 0
+    }
+
     /// Returns the current offset of the buffer.
     pub fn off(&self) -> usize {
         self.off
     }
 
     /// Returns a reference to the internal buffer.
-    pub fn buf(&self) -> &[u8] {
+    pub fn buf(&self) -> &'a [u8] {
         self.buf
     }
 
@@ -306,7 +378,7 @@ impl<'a> Octets<'a> {
     }
 }
 
-impl<'a> AsRef<[u8]> for Octets<'a> {
+impl AsRef<[u8]> for Octets<'_> {
     fn as_ref(&self) -> &[u8] {
         &self.buf[self.off..]
     }
@@ -315,7 +387,7 @@ impl<'a> AsRef<[u8]> for Octets<'a> {
 /// A zero-copy mutable byte buffer.
 ///
 /// Like `Octets` but mutable.
-#[derive(Debug, PartialEq)]
+#[derive(Debug, PartialEq, Eq)]
 pub struct OctetsMut<'a> {
     buf: &'a mut [u8],
     off: usize,
@@ -466,7 +538,7 @@ impl<'a> OctetsMut<'a> {
 
     /// Reads `len` bytes from the current offset without copying and advances
     /// the buffer.
-    pub fn get_bytes(&mut self, len: usize) -> Result<Octets> {
+    pub fn get_bytes(&mut self, len: usize) -> Result<Octets<'_>> {
         if self.cap() < len {
             return Err(BufferTooShortError);
         }
@@ -483,7 +555,7 @@ impl<'a> OctetsMut<'a> {
 
     /// Reads `len` bytes from the current offset without copying and advances
     /// the buffer.
-    pub fn get_bytes_mut(&mut self, len: usize) -> Result<OctetsMut> {
+    pub fn get_bytes_mut(&mut self, len: usize) -> Result<OctetsMut<'_>> {
         if self.cap() < len {
             return Err(BufferTooShortError);
         }
@@ -500,7 +572,7 @@ impl<'a> OctetsMut<'a> {
 
     /// Reads `len` bytes from the current offset without copying and advances
     /// the buffer, where `len` is an unsigned 8-bit integer prefix.
-    pub fn get_bytes_with_u8_length(&mut self) -> Result<Octets> {
+    pub fn get_bytes_with_u8_length(&mut self) -> Result<Octets<'_>> {
         let len = self.get_u8()?;
         self.get_bytes(len as usize)
     }
@@ -508,7 +580,7 @@ impl<'a> OctetsMut<'a> {
     /// Reads `len` bytes from the current offset without copying and advances
     /// the buffer, where `len` is an unsigned 16-bit integer prefix in network
     /// byte-order.
-    pub fn get_bytes_with_u16_length(&mut self) -> Result<Octets> {
+    pub fn get_bytes_with_u16_length(&mut self) -> Result<Octets<'_>> {
         let len = self.get_u16()?;
         self.get_bytes(len as usize)
     }
@@ -516,14 +588,14 @@ impl<'a> OctetsMut<'a> {
     /// Reads `len` bytes from the current offset without copying and advances
     /// the buffer, where `len` is an unsigned variable-length integer prefix
     /// in network byte-order.
-    pub fn get_bytes_with_varint_length(&mut self) -> Result<Octets> {
+    pub fn get_bytes_with_varint_length(&mut self) -> Result<Octets<'_>> {
         let len = self.get_varint()?;
         self.get_bytes(len as usize)
     }
 
     /// Reads `len` bytes from the current offset without copying and without
     /// advancing the buffer.
-    pub fn peek_bytes(&mut self, len: usize) -> Result<Octets> {
+    pub fn peek_bytes(&mut self, len: usize) -> Result<Octets<'_>> {
         if self.cap() < len {
             return Err(BufferTooShortError);
         }
@@ -538,7 +610,7 @@ impl<'a> OctetsMut<'a> {
 
     /// Reads `len` bytes from the current offset without copying and without
     /// advancing the buffer.
-    pub fn peek_bytes_mut(&mut self, len: usize) -> Result<OctetsMut> {
+    pub fn peek_bytes_mut(&mut self, len: usize) -> Result<OctetsMut<'_>> {
         if self.cap() < len {
             return Err(BufferTooShortError);
         }
@@ -551,8 +623,7 @@ impl<'a> OctetsMut<'a> {
         Ok(out)
     }
 
-    /// Writes `len` bytes from the current offset without copying and advances
-    /// the buffer.
+    /// Writes `v` to the current offset.
     pub fn put_bytes(&mut self, v: &[u8]) -> Result<()> {
         let len = v.len();
 
@@ -571,8 +642,83 @@ impl<'a> OctetsMut<'a> {
         Ok(())
     }
 
+    /// Writes `v` to the current offset after Huffman-encoding it.
+    ///
+    /// The Huffman code implemented is the one defined for HPACK (RFC7541).
+    #[cfg(feature = "huffman_hpack")]
+    pub fn put_huffman_encoded<const LOWER_CASE: bool>(
+        &mut self, v: &[u8],
+    ) -> Result<()> {
+        use self::huffman_table::ENCODE_TABLE;
+
+        let mut bits: u64 = 0;
+        let mut pending = 0;
+
+        for &b in v {
+            let b = if LOWER_CASE {
+                b.to_ascii_lowercase()
+            } else {
+                b
+            };
+            let (nbits, code) = ENCODE_TABLE[b as usize];
+
+            pending += nbits;
+
+            if pending < 64 {
+                // Have room for the new token
+                bits |= code << (64 - pending);
+                continue;
+            }
+
+            pending -= 64;
+            // Take only the bits that fit
+            bits |= code >> pending;
+            self.put_u64(bits)?;
+
+            bits = if pending == 0 {
+                0
+            } else {
+                code << (64 - pending)
+            };
+        }
+
+        if pending == 0 {
+            return Ok(());
+        }
+
+        bits |= u64::MAX >> pending;
+        // TODO: replace with `next_multiple_of(8)` when stable
+        pending = (pending + 7) & !7; // Round up to a byte
+        bits >>= 64 - pending;
+
+        if pending >= 32 {
+            pending -= 32;
+            self.put_u32((bits >> pending) as u32)?;
+        }
+
+        while pending > 0 {
+            pending -= 8;
+            self.put_u8((bits >> pending) as u8)?;
+        }
+
+        Ok(())
+    }
+
+    /// Rewinds the buffer offset by `len` elements.
+    pub fn rewind(&mut self, len: usize) -> Result<()> {
+        if self.off() < len {
+            return Err(BufferTooShortError);
+        }
+
+        self.off -= len;
+
+        Ok(())
+    }
+
     /// Splits the buffer in two at the given absolute offset.
-    pub fn split_at(&mut self, off: usize) -> Result<(OctetsMut, OctetsMut)> {
+    pub fn split_at(
+        &mut self, off: usize,
+    ) -> Result<(OctetsMut<'_>, OctetsMut<'_>)> {
         if self.len() < off {
             return Err(BufferTooShortError);
         }
@@ -626,6 +772,11 @@ impl<'a> OctetsMut<'a> {
         self.buf.len()
     }
 
+    /// Returns `true` if the buffer is empty.
+    pub fn is_empty(&self) -> bool {
+        self.buf.len() == 0
+    }
+
     /// Returns the current offset of the buffer.
     pub fn off(&self) -> usize {
         self.off
@@ -642,13 +793,13 @@ impl<'a> OctetsMut<'a> {
     }
 }
 
-impl<'a> AsRef<[u8]> for OctetsMut<'a> {
+impl AsRef<[u8]> for OctetsMut<'_> {
     fn as_ref(&self) -> &[u8] {
         &self.buf[self.off..]
     }
 }
 
-impl<'a> AsMut<[u8]> for OctetsMut<'a> {
+impl AsMut<[u8]> for OctetsMut<'_> {
     fn as_mut(&mut self) -> &mut [u8] {
         &mut self.buf[self.off..]
     }
@@ -656,14 +807,14 @@ impl<'a> AsMut<[u8]> for OctetsMut<'a> {
 
 /// Returns how many bytes it would take to encode `v` as a variable-length
 /// integer.
-pub fn varint_len(v: u64) -> usize {
+pub const fn varint_len(v: u64) -> usize {
     if v <= 63 {
         1
     } else if v <= 16383 {
         2
     } else if v <= 1_073_741_823 {
         4
-    } else if v <= 4_611_686_018_427_387_903 {
+    } else if v <= MAX_VAR_INT {
         8
     } else {
         unreachable!()
@@ -671,7 +822,7 @@ pub fn varint_len(v: u64) -> usize {
 }
 
 /// Returns how long the variable-length integer is, given its first byte.
-pub fn varint_parse_len(first: u8) -> usize {
+pub const fn varint_parse_len(first: u8) -> usize {
     match first >> 6 {
         0 => 1,
         1 => 2,
@@ -681,6 +832,58 @@ pub fn varint_parse_len(first: u8) -> usize {
     }
 }
 
+/// Returns how long the Huffman encoding of the given buffer will be.
+///
+/// The Huffman code implemented is the one defined for HPACK (RFC7541).
+#[cfg(feature = "huffman_hpack")]
+pub fn huffman_encoding_len<const LOWER_CASE: bool>(src: &[u8]) -> Result<usize> {
+    use self::huffman_table::ENCODE_TABLE;
+
+    let mut bits: usize = 0;
+
+    for &b in src {
+        let b = if LOWER_CASE {
+            b.to_ascii_lowercase()
+        } else {
+            b
+        };
+
+        let (nbits, _) = ENCODE_TABLE[b as usize];
+        bits += nbits;
+    }
+
+    let mut len = bits / 8;
+
+    if bits & 7 != 0 {
+        len += 1;
+    }
+
+    if len > src.len() {
+        return Err(BufferTooShortError);
+    }
+
+    Ok(len)
+}
+
+/// The functions in this mod test the compile time assertions in the
+/// `put_u` and `peek_u` macros. If you compile this crate with
+/// `--cfg test_invalid_len_compilation_fail`, e.g., by using
+/// `cargo rustc  -- --cfg test_invalid_len_compilation_fail`
+/// You will get two compiler errors
+#[cfg(test_invalid_len_compilation_fail)]
+pub mod fails_to_compile {
+    use super::*;
+    pub fn peek_invalid_fails_to_compile(b: &mut Octets) -> Result<u8> {
+        peek_u!(b, u8, 2)
+    }
+
+    pub fn put_invalid_fails_to_compile<'a>(
+        b: &'a mut OctetsMut, v: u8,
+    ) -> Result<&'a mut [u8]> {
+        put_u!(b, u8, v, 2)
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -955,6 +1158,27 @@ mod tests {
         assert_eq!(b.off(), 1);
     }
 
+    #[cfg(feature = "huffman_hpack")]
+    #[test]
+    fn invalid_huffman() {
+        // Extra non-zero padding byte at the end.
+        let mut b = Octets::with_slice(
+            b"\x00\x85\xf2\xb2\x4a\x84\xff\x84\x49\x50\x9f\xff",
+        );
+        assert!(b.get_huffman_decoded().is_err());
+
+        // Zero padded.
+        let mut b =
+            Octets::with_slice(b"\x00\x85\xf2\xb2\x4a\x84\xff\x83\x49\x50\x90");
+        assert!(b.get_huffman_decoded().is_err());
+
+        // Non-final EOS symbol.
+        let mut b = Octets::with_slice(
+            b"\x00\x85\xf2\xb2\x4a\x84\xff\x87\x49\x51\xff\xff\xff\xfa\x7f",
+        );
+        assert!(b.get_huffman_decoded().is_err());
+    }
+
     #[test]
     fn put_varint() {
         let mut d = [0; 8];
@@ -1013,7 +1237,7 @@ mod tests {
     fn varint_too_large() {
         let mut d = [0; 3];
         let mut b = OctetsMut::with_slice(&mut d);
-        assert!(b.put_varint(std::u64::MAX).is_err());
+        assert!(b.put_varint(u64::MAX).is_err());
     }
 
     #[test]
@@ -1075,6 +1299,56 @@ mod tests {
         assert_eq!(&d, &exp);
     }
 
+    #[test]
+    fn rewind() {
+        let d = [0xc2, 0x19, 0x7c, 0x5e, 0xff, 0x14, 0xe8, 0x8c];
+        let mut b = Octets::with_slice(&d);
+        assert_eq!(b.get_varint().unwrap(), 151288809941952652);
+        assert_eq!(b.cap(), 0);
+        assert_eq!(b.off(), 8);
+
+        assert_eq!(b.rewind(4), Ok(()));
+        assert_eq!(b.cap(), 4);
+        assert_eq!(b.off(), 4);
+
+        assert_eq!(b.get_u8().unwrap(), 0xff);
+        assert_eq!(b.cap(), 3);
+        assert_eq!(b.off(), 5);
+
+        assert!(b.rewind(6).is_err());
+
+        assert_eq!(b.rewind(5), Ok(()));
+        assert_eq!(b.cap(), 8);
+        assert_eq!(b.off(), 0);
+
+        assert!(b.rewind(1).is_err());
+    }
+
+    #[test]
+    fn rewind_mut() {
+        let mut d = [0xc2, 0x19, 0x7c, 0x5e, 0xff, 0x14, 0xe8, 0x8c];
+        let mut b = OctetsMut::with_slice(&mut d);
+        assert_eq!(b.get_varint().unwrap(), 151288809941952652);
+        assert_eq!(b.cap(), 0);
+        assert_eq!(b.off(), 8);
+
+        assert_eq!(b.rewind(4), Ok(()));
+        assert_eq!(b.cap(), 4);
+        assert_eq!(b.off(), 4);
+
+        assert_eq!(b.get_u8().unwrap(), 0xff);
+        assert_eq!(b.cap(), 3);
+        assert_eq!(b.off(), 5);
+
+        assert!(b.rewind(6).is_err());
+
+        assert_eq!(b.rewind(5), Ok(()));
+        assert_eq!(b.cap(), 8);
+        assert_eq!(b.off(), 0);
+
+        assert!(b.rewind(1).is_err());
+    }
+
     #[test]
     fn split() {
         let mut d = b"helloworld".to_vec();
@@ -1272,3 +1546,6 @@ mod tests {
         }
     }
 }
+
+#[cfg(feature = "huffman_hpack")]
+mod huffman_table;
